@@ -25,10 +25,6 @@
 #include "soup-dns.h"
 #include "soup-misc.h"
 
-#ifndef socklen_t
-#  define socklen_t int
-#endif
-
 #ifndef INET_ADDRSTRLEN
 #  define INET_ADDRSTRLEN 16
 #  define INET6_ADDRSTRLEN 46
@@ -617,9 +613,9 @@ check_hostent (SoupDNSEntry *entry, gboolean block)
 	}
 
 	if (block)
-		tvp = &tv;
-	else
 		tvp = NULL;
+	else
+		tvp = &tv;
 
 	do {
 		FD_ZERO (&readfds);
