@@ -3,6 +3,7 @@
  * Copyright (C) 2003, Novell, Inc.
  */
 
+#include <stdlib.h>
 #include <string.h>
 #include <libxml/tree.h>
 #include "soup-misc.h"
@@ -240,6 +241,17 @@ soup_soap_parameter_get_name (SoupSoapParameter *param)
 	return (const char *) param->name;
 }
 
+/**
+ * soup_soap_parameter_get_int_value:
+ */
+int
+soup_soap_parameter_get_int_value (SoupSoapParameter *param)
+{
+	g_return_val_if_fail (param != NULL, -1);
+
+	return atoi (xmlNodeGetContent (param));
+}
+
 /**
  * soup_soap_parameter_get_string_value:
  */
