@@ -66,6 +66,8 @@ struct SoupSocketPrivate {
 	guint           watch;
 	guint           read_tag, write_tag, error_tag;
 	GByteArray     *read_buf;
+
+	GMutex *iolock, *addrlock;
 };
 
 #ifdef HAVE_IPV6
@@ -88,6 +90,8 @@ init (GObject *object)
 	sock->priv->sockfd = -1;
 	sock->priv->non_blocking = sock->priv->nodelay = TRUE;
 	sock->priv->reuseaddr = TRUE;
+	sock->priv->addrlock = g_mutex_new ();
+	sock->priv->iolock = g_mutex_new ();
 }
 
 static void
@@ -95,6 +99,7 @@ disconnect_internal (SoupSocket *sock)
 {
 	g_io_channel_unref (sock->priv->iochannel);
 	sock->priv->iochannel = NULL;
+	sock->priv->sockfd = -1;
 
 	if (sock->priv->read_tag) {
 		g_source_remove (sock->priv->read_tag);
@@ -126,6 +131,9 @@ finalize (GObject *object)
 	if (sock->priv->watch)
 		g_source_remove (sock->priv->watch);
 
+	g_mutex_free (sock->priv->addrlock);
+	g_mutex_free (sock->priv->iolock);
+
 	g_free (sock->priv);
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -230,7 +238,7 @@ update_fdflags (SoupSocket *sock)
 {
 	int flags, opt;
 
-	if (!sock->priv->sockfd)
+	if (sock->priv->sockfd == -1)
 		return;
 
 	flags = fcntl (sock->priv->sockfd, F_GETFL, 0);
@@ -330,6 +338,7 @@ soup_socket_new (const char *optname1, ...)
 static GIOChannel *
 get_iochannel (SoupSocket *sock)
 {
+	g_mutex_lock (sock->priv->iolock);
 	if (!sock->priv->iochannel) {
 		sock->priv->iochannel =
 			g_io_channel_unix_new (sock->priv->sockfd);
@@ -337,6 +346,7 @@ get_iochannel (SoupSocket *sock)
 		g_io_channel_set_encoding (sock->priv->iochannel, NULL, NULL);
 		g_io_channel_set_buffered (sock->priv->iochannel, FALSE);
 	}
+	g_mutex_unlock (sock->priv->iolock);
 	return sock->priv->iochannel;
 }
 
@@ -471,9 +481,11 @@ soup_socket_connect (SoupSocket *sock, SoupAddress *remote_addr)
 	if (sock->priv->non_blocking) {
 		sock->priv->watch = g_idle_add (idle_connect_result, sock);
 		return SOUP_STATUS_CONTINUE;
-	} else {
-		return sock->priv->sockfd != -1 ?
-			SOUP_STATUS_OK : SOUP_STATUS_CANT_CONNECT;
+	} else if (sock->priv->sockfd == -1)
+		return SOUP_STATUS_CANT_CONNECT;
+	else {
+		get_iochannel (sock);
+		return SOUP_STATUS_OK;
 	}
 }
 
@@ -721,12 +733,37 @@ soup_socket_server_new (SoupAddress *local_addr, gpointer ssl_creds,
 void
 soup_socket_disconnect (SoupSocket *sock)
 {
+	gboolean already_disconnected = FALSE;
+
 	g_return_if_fail (SOUP_IS_SOCKET (sock));
 
-	if (!sock->priv->iochannel)
-		return;
+	if (g_mutex_trylock (sock->priv->iolock)) {
+		if (sock->priv->iochannel)
+			disconnect_internal (sock);
+		else
+			already_disconnected = TRUE;
+		g_mutex_unlock (sock->priv->iolock);
+	} else {
+		int sockfd;
 
-	disconnect_internal (sock);
+		/* Another thread is currently doing IO, so
+		 * we can't close the iochannel. So just kick
+		 * the file descriptor out from under it.
+		 */
+
+		sockfd = sock->priv->sockfd;
+		sock->priv->sockfd = -1;
+		if (sockfd == -1)
+			already_disconnected = TRUE;
+		else {
+			g_io_channel_set_close_on_unref (sock->priv->iochannel,
+							 FALSE);
+			close (sockfd);
+		}
+	}
+
+	if (already_disconnected)
+		return;
 
 	/* Give all readers a chance to notice the connection close */
 	g_signal_emit (sock, signals[READABLE], 0);
@@ -751,6 +788,7 @@ soup_socket_get_local_address (SoupSocket *sock)
 {
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
 
+	g_mutex_lock (sock->priv->addrlock);
 	if (!sock->priv->local_addr) {
 		struct soup_sockaddr_max bound_sa;
 		int sa_len;
@@ -759,6 +797,7 @@ soup_socket_get_local_address (SoupSocket *sock)
 		getsockname (sock->priv->sockfd, (struct sockaddr *)&bound_sa, &sa_len);
 		sock->priv->local_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&bound_sa, sa_len);
 	}
+	g_mutex_unlock (sock->priv->addrlock);
 
 	return sock->priv->local_addr;
 }
@@ -768,7 +807,8 @@ soup_socket_get_remote_address (SoupSocket *sock)
 {
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
 
-	if (!sock->priv->local_addr) {
+	g_mutex_lock (sock->priv->addrlock);
+	if (!sock->priv->remote_addr) {
 		struct soup_sockaddr_max bound_sa;
 		int sa_len;
 
@@ -776,6 +816,7 @@ soup_socket_get_remote_address (SoupSocket *sock)
 		getpeername (sock->priv->sockfd, (struct sockaddr *)&bound_sa, &sa_len);
 		sock->priv->remote_addr = soup_address_new_from_sockaddr ((struct sockaddr *)&bound_sa, sa_len);
 	}
+	g_mutex_unlock (sock->priv->addrlock);
 
 	return sock->priv->remote_addr;
 }
@@ -799,7 +840,7 @@ read_from_network (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 {
 	GIOStatus status;
 
-	if (!sock->priv->iochannel)
+	if (!sock->priv->iochannel) 
 		return SOUP_SOCKET_EOF;
 
 	status = g_io_channel_read_chars (sock->priv->iochannel,
@@ -871,12 +912,18 @@ read_from_buf (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 SoupSocketIOStatus
 soup_socket_read (SoupSocket *sock, gpointer buffer, gsize len, gsize *nread)
 {
+	SoupSocketIOStatus status;
+
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_SOCKET_ERROR);
 
+	g_mutex_lock (sock->priv->iolock);
 	if (sock->priv->read_buf)
-		return read_from_buf (sock, buffer, len, nread);
+		status = read_from_buf (sock, buffer, len, nread);
 	else
-		return read_from_network (sock, buffer, len, nread);
+		status = read_from_network (sock, buffer, len, nread);
+	g_mutex_unlock (sock->priv->iolock);
+
+	return status;
 }
 
 /**
@@ -910,6 +957,8 @@ soup_socket_read_until (SoupSocket *sock, gpointer buffer, gsize len,
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_SOCKET_ERROR);
 	g_return_val_if_fail (len >= boundary_len, SOUP_SOCKET_ERROR);
 
+	g_mutex_lock (sock->priv->iolock);
+
 	*got_boundary = FALSE;
 
 	if (!sock->priv->read_buf)
@@ -924,8 +973,10 @@ soup_socket_read_until (SoupSocket *sock, gpointer buffer, gsize len,
 					    len - prev_len, nread);
 		read_buf->len = prev_len + *nread;
 
-		if (status != SOUP_SOCKET_OK)
+		if (status != SOUP_SOCKET_OK) {
+			g_mutex_unlock (sock->priv->iolock);
 			return status;
+		}
 	}
 
 	/* Scan for the boundary */
@@ -943,7 +994,10 @@ soup_socket_read_until (SoupSocket *sock, gpointer buffer, gsize len,
 	 * buffer).
 	 */
 	match_len = p - read_buf->data;
-	return read_from_buf (sock, buffer, MIN (len, match_len), nread);
+	status = read_from_buf (sock, buffer, MIN (len, match_len), nread);
+
+	g_mutex_unlock (sock->priv->iolock);
+	return status;
 }
 
 static gboolean
@@ -987,23 +1041,34 @@ soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 
 	g_return_val_if_fail (SOUP_IS_SOCKET (sock), SOUP_SOCKET_ERROR);
 
-	if (!sock->priv->iochannel)
+	g_mutex_lock (sock->priv->iolock);
+
+	if (!sock->priv->iochannel) {
+		g_mutex_unlock (sock->priv->iolock);
 		return SOUP_SOCKET_EOF;
-	if (sock->priv->write_tag)
+	}
+	if (sock->priv->write_tag) {
+		g_mutex_unlock (sock->priv->iolock);
 		return SOUP_SOCKET_WOULD_BLOCK;
+	}
 
 	pipe_handler = signal (SIGPIPE, SIG_IGN);
 	status = g_io_channel_write_chars (sock->priv->iochannel,
 					   buffer, len, nwrote, NULL);
 	signal (SIGPIPE, pipe_handler);
-	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN)
+	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN) {
+		g_mutex_unlock (sock->priv->iolock);
 		return SOUP_SOCKET_ERROR;
+	}
 
-	if (*nwrote)
+	if (*nwrote) {
+		g_mutex_unlock (sock->priv->iolock);
 		return SOUP_SOCKET_OK;
+	}
 
 	sock->priv->write_tag =
 		g_io_add_watch (sock->priv->iochannel, G_IO_OUT,
 				socket_write_watch, sock);
+	g_mutex_unlock (sock->priv->iolock);
 	return SOUP_SOCKET_WOULD_BLOCK;
 }
